package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntSet
import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.items.IItemHandler
import ru.dbotthepony.kommons.collect.any
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.map

/**
 * Container which revolve around embedding slot objects rather than providing direct item access,
 * and subsequently fully implement [IItemHandler]
 */
interface ISlottedContainer : IEnhancedContainer, IItemHandler {
	override fun containerSlot(slot: Int): ISlottedContainerSlot

	override fun slotIterator(): Iterator<ISlottedContainerSlot> {
		return (0 until containerSize).iterator().map { containerSlot(it) }
	}

	override fun nonEmptySlotIterator(): Iterator<ISlottedContainerSlot> {
		return (0 until containerSize).iterator().map { containerSlot(it) }.filter { it.isNotEmpty }
	}

	override fun setChanged(slot: Int) {
		containerSlot(slot).setChanged()
	}

	override fun getMaxStackSize(slot: Int, itemStack: ItemStack): Int {
		return containerSlot(slot).maxStackSize(itemStack)
	}

	override fun getItem(slot: Int): ItemStack {
		return containerSlot(slot).item
	}

	override fun removeItem(slot: Int, amount: Int): ItemStack {
		return containerSlot(slot).remove(amount)
	}

	override fun removeItemNoUpdate(slot: Int): ItemStack {
		return containerSlot(slot).remove()
	}

	override fun setItem(slot: Int, itemStack: ItemStack) {
		containerSlot(slot).item = itemStack
	}

	override fun canPlaceItem(slot: Int, itemStack: ItemStack): Boolean {
		return containerSlot(slot).canAutomationPlaceItem(itemStack)
	}

	override fun canTakeItem(container: Container, slot: Int, itemStack: ItemStack): Boolean {
		return containerSlot(slot).canAutomationTakeItem()
	}

	override fun getSlots() = containerSize
	override fun getStackInSlot(slot: Int) = containerSlot(slot).item

	override fun insertItem(slot: Int, stack: ItemStack, simulate: Boolean): ItemStack {
		return containerSlot(slot).insertItem(stack, simulate)
	}

	override fun extractItem(slot: Int, amount: Int, simulate: Boolean): ItemStack {
		return containerSlot(slot).extractItem(amount, simulate)
	}

	override fun getSlotLimit(slot: Int): Int {
		return containerSlot(slot).maxStackSize
	}

	override fun isItemValid(slot: Int, stack: ItemStack): Boolean {
		return canPlaceItem(slot, stack)
	}

	private fun addItem(stack: ItemStack, simulate: Boolean, filterPass: Boolean, slots: IntSet, onlyIntoExisting: Boolean, popTime: Int?, ignoreFilters: Boolean): ItemStack {
		if (stack.isEmpty || slots.isEmpty())
			return stack

		// двигаем в одинаковые слоты
		for (i in slotWithItemIterator(stack.item, slots)) {
			val slot = containerSlot(i)

			val condition: Boolean

			if (slot is IFilteredContainerSlot) {
				condition = (ignoreFilters || !slot.isForbiddenForAutomation) &&
					ItemStack.isSameItemSameComponents(slot.item, stack) &&
					(ignoreFilters || !filterPass && !slot.hasFilter || filterPass && slot.hasFilter && slot.testSlotFilter(stack))
			} else {
				condition = (ignoreFilters || !filterPass) && ItemStack.isSameItemSameComponents(slot.item, stack)
			}

			if (condition) {
				val slotLimit = slot.maxStackSize(slot.item)

				if (slot.item.count < slotLimit) {
					val newCount = (slot.item.count + stack.count).coerceAtMost(slotLimit)
					val diff = newCount - slot.item.count

					if (!simulate) {
						slot.item.count = newCount
						slot.setChanged()

						if (popTime != null) {
							slot.item.popTime = popTime
						}
					}

					stack.shrink(diff)

					if (stack.isEmpty)
						return ItemStack.EMPTY
				}
			}
		}

		if (!onlyIntoExisting) {
			for (i in emptySlotIterator(slots)) {
				val slot = containerSlot(i)

				val condition: Boolean

				if (slot is IFilteredContainerSlot) {
					condition = (ignoreFilters || !slot.isForbiddenForAutomation) &&
						(ignoreFilters || !filterPass && !slot.hasFilter || filterPass && slot.hasFilter && slot.testSlotFilter(stack))
				} else {
					condition = ignoreFilters || !filterPass
				}

				if (condition) {
					val diff = stack.count.coerceAtMost(slot.maxStackSize(stack))

					if (!simulate) {
						val copyToPut = stack.copy()
						copyToPut.count = diff
						slot.item = copyToPut

						if (popTime != null) {
							copyToPut.popTime = popTime
						}
					}

					stack.shrink(diff)

					if (stack.isEmpty)
						return ItemStack.EMPTY
				}
			}
		}

		return stack
	}

	/**
	 * Hint used internally by [ISlottedContainer] to potentially speed up default method implementations
	 */
	val hasFilterableSlots: Boolean
		get() = slotIterator().any { it is IFilteredContainerSlot }

	fun addItem(stack: ItemStack, simulate: Boolean, slots: IntSet = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean): ItemStack {
		if (stack.isEmpty || slots.isEmpty())
			return stack

		if (ignoreFilters || !hasFilterableSlots) {
			return addItem(stack.copy(), simulate, filterPass = true, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = true)
		} else {
			var copy = addItem(stack.copy(), simulate, filterPass = true, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = false)
			copy = addItem(copy, simulate, filterPass = false, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = false)
			return copy
		}
	}


	/**
	 * Unlike [addItem], modifies original [stack]
	 *
	 * @return Whenever [stack] was modified
	 */
	fun consumeItem(stack: ItemStack, simulate: Boolean, slots: IntSet = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean): Boolean {
		if (stack.isEmpty)
			return false

		val result = addItem(stack, simulate, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = ignoreFilters)
		if (!simulate) stack.count = result.count
		return result.count != stack.count
	}

	fun fullyAddItem(stack: ItemStack, slots: IntSet = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean): Boolean {
		if (!addItem(stack, true, slots, popTime = popTime, onlyIntoExisting = onlyIntoExisting, ignoreFilters = ignoreFilters).isEmpty)
			return false

		return addItem(stack, false, slots, popTime = popTime, onlyIntoExisting = onlyIntoExisting, ignoreFilters = ignoreFilters).isEmpty
	}

	override fun addItem(
		stack: ItemStack,
		simulate: Boolean,
		slots: IntSet,
		onlyIntoExisting: Boolean,
		popTime: Int?
	): ItemStack {
		return addItem(stack, simulate, slots, onlyIntoExisting, popTime, ignoreFilters = false)
	}

	override fun consumeItem(
		stack: ItemStack,
		simulate: Boolean,
		slots: IntSet,
		onlyIntoExisting: Boolean,
		popTime: Int?
	): Boolean {
		return consumeItem(stack, simulate, slots, onlyIntoExisting, popTime, ignoreFilters = false)
	}

	override fun fullyAddItem(stack: ItemStack, slots: IntSet, onlyIntoExisting: Boolean, popTime: Int?): Boolean {
		return fullyAddItem(stack, slots, onlyIntoExisting, popTime, ignoreFilters = false)
	}
}
